import json

from backend.models import View, Keyword, FedCard, DssUser, Trigger, Connection, Card
import backend.controllers as ctrls
from backend.schemas import ViewSchema, NotProvided, KeywordSchema
from werkzeug.exceptions import NotFound, BadRequest
from marshmallow.exceptions import ValidationError
from typing import Any, List, Dict, Mapping, Tuple, Union, NoReturn
import hashlib


class ViewController(ctrls.BaseController):
    def get_all(self) -> List[View]:
        return View.query.all()

    def get_by_id(self, view_id: int, ignore_if_not_found: bool = False) -> Union[View, NoReturn]:
        if view_id is None:
            if not ignore_if_not_found:
                raise ValidationError(message={"view_id": ["Field may not be null."]})
            return None
        view = View.query.get(view_id)
        if view is None and not ignore_if_not_found:
            raise NotFound("View not found.")
        return view

    def get_view_connection(self, view_id: int) -> Connection:
        view = self.get_by_id(view_id)
        view_connection = ctrls.connection_ctrl.get_by_id_and_set_default_keyword_if_needed(view.connection.id)
        return view_connection

    def get_view_cards(self, view_id: int) -> List[Card]:
        view = self.get_by_id(view_id)
        cards = sorted(view.cards, key=lambda c: c.priority)
        return cards

    def get_all_active(self) -> List[View]:
        all_views = self.get_all()
        return [v for v in all_views if self.is_active(v.id)]

    def import_(
        self,
        name: str,
        connection: dict,
        cards: List[dict],
        url_rules: List[dict],
        description: str = NotProvided,
    ) -> View:
        new_view = ViewSchema().load(
            {
                "name": name,
                "description": description or "",
            },
            session=self.db.session,
        )
        self.db.session.add(new_view)
        self.db.session.commit()

        ctrls.card_ctrl.create_many(view_id=new_view.id, cards=cards)
        ctrls.url_rule_ctrl.create_many(view_id=new_view.id, url_rules=url_rules)
        ctrls.connection_ctrl.create(view_id=new_view.id, **connection)

        return self.get_by_id(new_view.id)

    def export(self, view_id: int) -> Dict:
        view = self.get_by_id(view_id)

        fields_to_export = ["name", "description"]

        view_as_dict = ViewSchema(only=tuple(fields_to_export)).dump(view)
        view_as_dict["cards"] = [ctrls.card_ctrl.export(card.id) for card in view.cards]
        view_as_dict["url_rules"] = [ctrls.url_rule_ctrl.export(r.id) for r in view.url_rules]
        view_as_dict["connection"] = ctrls.connection_ctrl.export(view.connection.id)

        export_name_prefix = " - Copy"
        view_as_dict["name"] = f"{view_as_dict['name']}{export_name_prefix}"
        return view_as_dict

    def feed_cards(self, view_id: int, keyword: Mapping[str, Any], dss_user: DssUser) -> List[FedCard]:
        view = self.get_by_id(view_id)
        keyword = KeywordSchema().load(keyword)
        if not view.is_active:
            return []

        fed_cards = []

        connection_data = ctrls.connection_ctrl.fetch_data(view.connection.id, keyword=keyword)

        for card in view.cards:
            fed_card = ctrls.card_ctrl.feed_one(card_id=card.id, connection_data=connection_data, dss_user=dss_user)
            if fed_card.data:
                fed_cards.append(fed_card)

        ordered_fed_cards = self.order_fed_cards(view_id, fed_cards, dss_user)
        return ordered_fed_cards

    def order_fed_cards(self, view_id: int, fed_cards: List[FedCard], dss_user: DssUser) -> List[FedCard]:
        def sorting_key(fed_card: FedCard):
            if fed_card.card.id in fed_cards_ordering:
                return fed_cards_ordering.index(fed_card.card.id)
            else:
                return fed_card.card.priority

        fed_cards_ordering = ctrls.user_config_ctrl.get_cards_ordering_for_view(view_id, dss_user=dss_user)
        ordered_fed_cards = sorted(fed_cards, key=sorting_key)
        for i, card in enumerate(ordered_fed_cards):
            card.position = i
        return ordered_fed_cards

    def create(self, name: str, description: str = NotProvided) -> View:
        new_view = ViewSchema().load(
            {
                "name": name,
                "description": description or "",
                "is_active": True,
            },
            session=self.db.session,
        )

        self.db.session.add(new_view)
        self.db.session.commit()

        ctrls.connection_ctrl.create(new_view.id)

        return self.get_by_id(new_view.id)

    def update(
        self,
        view_id: int,
        name: str = NotProvided,
        description: str = NotProvided,
        is_active: str = NotProvided,
    ) -> View:
        view = self.get_by_id(view_id)

        new_view = ViewSchema().load(
            {"name": name, "is_active": is_active, "description": description},
            instance=view,
            session=self.db.session,
            partial=("id"),
        )

        self.db.session.add(new_view)
        self.db.session.commit()

        return self.get_by_id(view_id)

    def delete(self, view_id: int) -> bool:
        view_to_delete = self.get_by_id(view_id)

        self.db.session.delete(view_to_delete)
        self.db.session.commit()

        is_view_deleted = self.get_by_id(view_id, ignore_if_not_found=True) == None
        if not is_view_deleted:
            raise BadRequest("Could not delete view.")

        return True

    def fetch_all_keywords(self, view_id: int) -> List[Keyword]:
        view = self.get_by_id(view_id)

        all_keywords = ctrls.connection_ctrl.fetch_keywords(connection_id=view.connection.id)

        return all_keywords

    def is_active(self, view_id: int) -> bool:
        view = self.get_by_id(view_id, ignore_if_not_found=True)
        if view is None:
            return False
        return view.is_active

    def get_url_rules(self, view_id: int):
        _ = self.get_by_id(view_id)

        url_rules = ctrls.url_rule_ctrl.get_all_for_view(view_id=view_id)
        return url_rules

    def generate_trigger(self, view_id: int) -> Trigger:
        view = ctrls.view_ctrl.get_by_id(view_id)
        view_keywords = ctrls.connection_ctrl.fetch_keywords(view.connection.id)
        view_refresh_hash = self.generate_refresh_hash(view_id=view_id)
        fed_url_rules = self.feed_url_rules(view_id=view_id)

        return Trigger(
            view=view,
            cards_count=len(view.cards),
            keywords=view_keywords,
            refresh_hash=view_refresh_hash,
            fed_url_rules=fed_url_rules,
        )

    def generate_many_triggers(self, views_id: List[int]) -> List[Trigger]:
        return [self.generate_trigger(view_id) for view_id in views_id]

    def generate_refresh_hash(self, view_id: int) -> str:
        view = ctrls.view_ctrl.get_by_id(view_id)
        connection = view.connection

        refresh_data = {
            "dataset_name": connection.dataset_name,
            "keywords_column_name": connection.keywords_column_name,
            "secondary_column_name": connection.secondary_column_name,
        }

        refresh_data["view"] = {"name": view.name, "active": view.is_active}

        url_rules = ctrls.view_ctrl.get_url_rules(connection.view_id)

        refresh_data["url_rules"] = [
            {
                "mode": r.mode,
                "matching_type": r.matching_type,
                "value": r.value,
                "priority": r.priority,
                "css_selectors": r.css_selectors,
            }
            for r in url_rules
        ]

        refresh_data_as_json = json.dumps(refresh_data)
        refresh_data_hash = hashlib.md5(refresh_data_as_json.encode("utf-8")).hexdigest()
        return refresh_data_hash

    def get_refreshed_and_active_triggers(self, refresh_hashes: Dict[str, str], force=False) -> Tuple:
        active_views_id = [view.id for view in self.get_all_active()]
        views_id_to_refresh = self.get_views_id_to_refresh(
            refresh_hashes=refresh_hashes, active_views_id=active_views_id, force=force
        )
        refreshed_triggers = self.generate_many_triggers(views_id_to_refresh)
        return refreshed_triggers, active_views_id

    def get_views_id_to_refresh(self, refresh_hashes: Dict[str, str], active_views_id: list, force=False):
        views_id_to_refresh = []
        for active_view_id in active_views_id:
            _active_view_id_cast = str(active_view_id)
            if _active_view_id_cast in list(refresh_hashes.keys()):
                current_view_hash = self.generate_refresh_hash(active_view_id)
                if refresh_hashes[_active_view_id_cast] != current_view_hash or force:
                    views_id_to_refresh.append(active_view_id)
            else:
                views_id_to_refresh.append(active_view_id)
        return views_id_to_refresh

    def feed_url_rules(self, view_id):
        _ = self.get_by_id(view_id)

        url_rules = ctrls.view_ctrl.get_url_rules(view_id)
        fed_url_rules = [ctrls.url_rule_ctrl.feed_url_rule(r.id) for r in url_rules]

        return fed_url_rules

    def get_connection_test_keyword(self, view_id: int) -> Keyword:
        view = self.get_by_id(view_id)

        test_keyword = ctrls.connection_ctrl.get_test_keyword(view.connection.id)

        return test_keyword

    def fetch_test_keyword_data(self, view_id: int) -> Dict:
        view = self.get_by_id(view_id)

        connection_data = ctrls.connection_ctrl.fetch_test_data(connection_id=view.connection.id)

        return connection_data
